import { IPlayable } from "../moq";
/**
 * This class sets rules when a configured setup should be applied to the target invocation or not.
 *
 * @example
 * ```typescript
 *
 * const value = 'value';
 * const object = new Mock<Function>()
 * .setup(instance => instance(1))
 * .play(PlayTimes.Once())
 * .returns(value)
 * .object();
 *
 * expect(object(1).toBe(value);
 * expect(object(1).toBe(undefined);
 * ```
 */
export declare class PlayTimes {
    private static readonly always;
    private static readonly never;
    /**
     * The configured setup will be applied to invocations exactly n-times. After that it will be ignored.
     */
    static Exactly(count: number): IPlayable;
    /**
     * The configured setup will be always applied to invocations.
     */
    static Always(): IPlayable;
    /**
     * The configured setup will be never applied to invocations.
     */
    static Never(): IPlayable;
    /**
     * The configured setup will be applied only to the first invocation.
     */
    static Once(): IPlayable;
    /**
     * The configured setup will be applied or not accordingly to the value in the sequence.
     *
     * @example
     * ```typescript
     *
     * const value = 'value';
     * const object = new Mock<Function>()
     * .setup(instance => instance(1))
     * .play(PlayTimes.Sequence([false, true]))
     * .returns(value)
     * .object();
     *
     * expect(object(1).toBe(undefined);
     * expect(object(1).toBe(value);
     * expect(object(1).toBe(undefined);
     * ```
     */
    static Sequence(sequence: boolean[]): IPlayable;
}
